#include "students_tab.hpp"

#include "data/data_manager.hpp"
#include "data/graph.hpp"
#include "data/nodes/dynexite_exam.hpp"
#include "util/enums.hpp"
#include "util/grade.hpp"
#include "util/points.hpp"
#include "util/strings.hpp"

#include <QBrush>
#include <QLabel>
#include <QList>
#include <QMessageBox>
#include <QScopeGuard>
#include <QString>
#include <QTableWidgetItem>
#include <QVBoxLayout>
#include <QVariant>

#include <array>
#include <cassert>
#include <initializer_list>
#include <ranges>

/*
 * Order of headers:
 *
 * - Student ID
 * - Closed headers (if there are closed points):
 *      - Points without Bonus (if there are bonus points)
 *      - Points with Bonus
 *      - Grade
 * - Open headers (if there are open points):
 *      - Points without Bonus (if there are bonus points)
 *      - Points with Bonus
 *      - Grade
 * - Combined grade (if there are closed and open points)
 */

namespace {
constexpr auto studentTableHeader = "Student";
constexpr auto combinedTableHeader = "Combined Grade";

constexpr EnumStorage<std::array<const char *, 3>> bonusTableHeadersPerType = {
    std::array{"Closed points (excl. Bonus)", "Closed points (incl. Bonus)",
               "Closed Grade"},
    std::array{"Open points (excl. Bonus)", "Open points (incl. Bonus)",
               "Open Grade"}};
constexpr EnumStorage<std::array<const char *, 2>> noBonusTableHeadersPerType =
    {std::array{"Closed points", "Closed Grade"},
     std::array{"Open points", "Open Grade"}};

/**
 * \brief Wait this amount of time after a resize is signaled to change table
 * column width and table row height.
 *
 * If we resize during the actual resizeEvent, the old window size is sometimes
 * used (especially when going from full screen to window-sized). The value of
 * this variable is chosen arbitrarily. In the future we might find a signal
 * that fires after a resize is completed, and can use that to resize the table,
 * but currently I don't know of any such signal.
 */
constexpr auto resizeWaitTime = std::chrono::milliseconds{100};

/**
 * \brief Calculates how many columns both \ref GradingType (closed or open)
 * occupy on their own.
 */
[[nodiscard]] int calcColumnsPerType(const DAG &cur) noexcept {
    return static_cast<int>(
        cur.hasBonus()
            ? std::tuple_size_v<decltype(bonusTableHeadersPerType)::value_type>
            : std::tuple_size_v<
                  decltype(noBonusTableHeadersPerType)::value_type>);
}

/**
 * \brief Calculates how many columns are currently needed in the table.
 */
[[nodiscard]] int calcTotalNrColumns(const DAG &cur) noexcept {
    using enum GradingType;
    const auto columnsPerType = calcColumnsPerType(cur);
    // 1 column for Student ID
    // calcColumnsPerType(...) columns for closed (if exist)
    // calcColumnsPerType(...) columns for open (if exist)
    // 1 column for combined (if open and closed exist)
    return 1 + columnsPerType * static_cast<int>(cur.hasType(Closed)) +
           columnsPerType * static_cast<int>(cur.hasType(Open)) +
           static_cast<int>(cur.hasType(Combined));
}

/**
 * \brief Calculates the indices of the grade columns for all \ref GradingType
 * \param cur Current exam
 *
 * If an exam does not use all grading types, the returned value at the index of
 * an unused grading type is invalid.
 */
[[nodiscard]] auto getGradeHeaderIndices(const DAG &cur) noexcept {
    const auto columnsPerType = calcColumnsPerType(cur);

    // The grade column is the last of the gradingType-specific columns.
    // Therefore the first grade column is the (1+columnsPerType)th column,
    // meaning the column at index columnsPerType.
    return getStorage(columnsPerType,
                      cur.hasType(GradingType::Closed) ? 2 * columnsPerType
                                                       : columnsPerType,
                      2 * columnsPerType + 1);
}

/**
 * \brief The background color of a grade cell changes depending on the grade.
 */
[[nodiscard]] QBrush getGradeBackgroundColor(const Grade g) noexcept {
    switch (g) {
    case Grade::grade1_0:
        return {Qt::green};
    case Grade::grade5_0:
        return {Qt::red};
    default:
        return {Qt::darkGreen};
    }
}

/**
 * \brief Inserts a new QTableWidgetItem into the table at the specified row and
 * column, using our preferred flags.
 */
void createTableItem(QTableWidget &table, const QVariant &value, const int row,
                     const int column) noexcept {
    const auto idItem = new QTableWidgetItem;
    idItem->setData(Qt::DisplayRole, value);
    idItem->setFlags(idItem->flags() ^ Qt::ItemIsEditable);
    table.setItem(row, column, idItem);
}

/**
 * \brief Create horizontal headers / columns.
 */
void createHeaders(QTableWidget &table, const DAG &cur) noexcept {
    using enum GradingType;
    table.setColumnCount(calcTotalNrColumns(cur));

    int columnIdx = 0;

    // Create the "student id"-header
    table.setHorizontalHeaderItem(columnIdx++,
                                  new QTableWidgetItem(studentTableHeader));

    // Create the point- and grade-headers for both closed and open
    for (const auto g : {Closed, Open}) {
        if (cur.hasType(g)) {

            // Create a header for each string in the range
            const auto createHeadersFromRange =
                [&table, g, &columnIdx](const std::ranges::range auto &range) {
                    for (const auto &s : range[g]) {
                        table.setHorizontalHeaderItem(columnIdx++,
                                                      new QTableWidgetItem(s));
                    }
                };

            if (cur.hasBonus()) {
                createHeadersFromRange(bonusTableHeadersPerType);
            } else {
                createHeadersFromRange(noBonusTableHeadersPerType);
            }
        }
    }

    // Create the "combined grade"-header if necessary
    if (cur.hasType(Combined)) {
        table.setHorizontalHeaderItem(
            columnIdx++, new QTableWidgetItem(combinedTableHeader));
    }
}

/**
 * \brief Create cells in each row, fill in Student ID and points (but not
 * grade).
 */
void insertData(QTableWidget &table, const DAG &cur) {
    using enum GradingType;

    // Sorting has to be disabled while inserting items
    table.setSortingEnabled(false);
    const auto enableSortingAgain =
        qScopeGuard([&table] { table.setSortingEnabled(true); });

    const auto &ids = cur.mExam.mIdentifier;
    const auto &pointsWithBonus = getStorage(cur.getPointsWithBonus(Closed),
                                             cur.getPointsWithBonus(Open));
    const auto &pointsWithoutBonus =
        getStorage(cur.mExam.mPointsWithoutBonus[Closed],
                   cur.mExam.mPointsWithoutBonus[Open]);

    table.setRowCount(static_cast<int>(ids.size()));
    assert(ids.size() == pointsWithoutBonus[Closed].size());
    assert(ids.size() == pointsWithoutBonus[Open].size());
    assert(ids.size() == pointsWithBonus[Closed].size());
    assert(ids.size() == pointsWithBonus[Open].size());

    // Write the student IDs, points and grades into each row
    for (qsizetype row = 0; row < ids.size(); ++row) {

        int columnIdx = 0;

        // Inserts the value into the current row and increments the column idx
        const auto createTableItemThisRow = [&table, row, &columnIdx](
                                                const QVariant &value) {
            createTableItem(table, value, static_cast<int>(row), columnIdx++);
        };

        // Student ID
        createTableItemThisRow(ids[row]);

        // Open and closed points and grades
        for (const auto g : {Closed, Open}) {
            if (cur.hasBonus()) {
                createTableItemThisRow(pointsWithoutBonus[g][row].p);
            }
            createTableItemThisRow(pointsWithBonus[g][row].p);

            // Grade (grade strings are filled in later)
            createTableItemThisRow("");
        }

        // Combined grade (grade strings are filled in later)
        createTableItemThisRow("");
    }
}
} // namespace

StudentsTab::StudentsTab(QWidget *parent) noexcept
    : QWidget(parent), mResizeTimer(this) {

    const auto layout = new QVBoxLayout(this);

    layout->addWidget(new QLabel(str::info::studentsTableTabLegend));
    layout->addWidget(&mTable);

    connect(&DataManager::getInstance(), &DataManager::rawDataChanged, this,
            &StudentsTab::pointsChanged);
    connect(&DataManager::getInstance(), &DataManager::gradesChanged, this,
            &StudentsTab::gradesChanged);

    mResizeTimer.callOnTimeout(this, &StudentsTab::resizeTable);
    // fire only once when timeouted, not repeatedly
    mResizeTimer.setSingleShot(true);
}

void StudentsTab::handleResize() noexcept {
    // Wait until resize is done and then fix column width and row height
    mResizeTimer.start(resizeWaitTime);
}

void StudentsTab::pointsChanged() noexcept {
    const auto cur = DataManager::getCurrent();

    // Clear content and headers
    mTable.clear();

    if (!cur) {
        return;
    }

    createHeaders(mTable, *cur);
    insertData(mTable, *cur);
}

void StudentsTab::gradesChanged(const EnumStorage<bool> &changedTypes) {

    using enum GradingType;

    const auto cur = DataManager::getCurrent();

    if (!cur) {
        return;
    }

    // Sorting has to be disabled while changing items
    mTable.setSortingEnabled(false);
    const auto enableSortingAgain =
        qScopeGuard([this] { mTable.setSortingEnabled(true); });

    const auto nrRows = mTable.rowCount();
    const auto gradeColumnNrs = getGradeHeaderIndices(*cur);

    // Store error strings
    QList<QString> notFoundIds;

    // Find corresponding grade to the student id in the row
    for (int i = 0; i < nrRows; ++i) {
        const auto id = mTable.item(i, 0)->text();

        const auto idx = cur->mExam.mIdentifier.indexOf(id);

        // Not found?
        if (idx == -1) {
            notFoundIds.push_back(id);
            continue;
        }

        // Set new grade strings and new background color
        // Use strings instead of doubles for a fixed number of decimal places
        for (const auto type : {Closed, Open, Combined}) {
            if (cur->hasType(type) && changedTypes[type]) {
                const auto item = mTable.item(i, gradeColumnNrs[type]);
                const auto grade = cur->getGradeList(type)[idx];
                item->setText(toString(grade));
                item->setBackground(getGradeBackgroundColor(grade));
            }
        }
    }

    // Display warnings if some occurred
    if (!notFoundIds.empty()) {
        assert(false);

        QString informativeText = "";
        for (const auto &id : notFoundIds) {
            informativeText += id + "\n";
        }

        QMessageBox msgBox(this);
        msgBox.setTextFormat(Qt::TextFormat::MarkdownText);
        msgBox.setWindowTitle("Students-Tab warning.");
        msgBox.setText("**Failed to look up grades for the following\nstudent "
                       "ids in the students tab:**\n");
        msgBox.setInformativeText(informativeText);
        msgBox.setStandardButtons(QMessageBox::Ok);
        msgBox.exec();
    }
}

void StudentsTab::resizeTable() noexcept {
    mTable.resizeColumnsToContents();
    mTable.resizeRowsToContents();
}
